# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from typing import ContextManager, Any

from rich.console import Console

from ..utils import get_rich_console, NO_RICH, dummy_context


def show_spinner(message: str, console: Console | None = None) -> ContextManager[Any]:
    """Context manager for showing a spinner during operations"""

    if console is None:
        console = get_rich_console()

    if bool(NO_RICH.value) or not sys.stdout.isatty():
        console.print(message)
        return dummy_context(None)

    return console.status(f"[cyan]{message}[/cyan]", spinner="dots")



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import textwrap

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich import box

from ..utils import get_rich_console
from ..models.openapi import (
    AppEvaluationEvaluationStepBenchmarkingResult,
    LogMessage,
    ProblemValidationTaskStatus,
    StepStatus,
)
from .strings import format_status, format_time, create_styled_table


def create_logs_table(
    logs: list[LogMessage] | None,
    title: str = "Validation Logs",
    max_message_len: int = 200,
) -> Table | None:
    if not logs:
        return None

    table = create_styled_table(title)
    table.add_column("Step", style="cyan", no_wrap=True)
    table.add_column("Status", no_wrap=True)
    table.add_column("Message", overflow="fold")

    for log in logs:
        status_display = format_status(log.type)
        message = log.message or ""
        if len(message) > max_message_len:
            message = message[: max_message_len - 3] + "..."
        table.add_row(log.step, status_display, message)

    return table


def print_logs(
    logs: list[LogMessage] | None,
    console: Console | None = None,
    title: str = "Validation Logs",
) -> None:
    if console is None:
        console = get_rich_console()
    table = create_logs_table(logs, title)
    if table:
        console.print(table)


def create_benchmark_table(
    result: AppEvaluationEvaluationStepBenchmarkingResult,
    title: str = "Benchmark Results",
) -> Table | None:
    if not result.benchmarked:
        return None

    table = create_styled_table(title)
    table.add_column("Metric", style="dim")
    table.add_column("Value", justify="right")

    if result.ref_time is not None:
        table.add_row("Reference (eager)", format_time(result.ref_time, result.ref_time_unit))
    if result.ref_compiled_time is not None:
        table.add_row("Reference (compiled)", format_time(result.ref_compiled_time, result.ref_compiled_time_unit))
    if result.optimized_time is not None:
        if isinstance(result.optimized_time, list):
            for i, t in enumerate(result.optimized_time):
                table.add_row(f"Optimized (shape {i + 1})", format_time(t, result.optimized_time_unit))
        else:
            table.add_row("Optimized", format_time(result.optimized_time, result.optimized_time_unit))

    return table


def print_benchmark_result(
    result: AppEvaluationEvaluationStepBenchmarkingResult | None,
    console: Console | None = None,
    title: str = "Benchmark Results",
) -> None:
    if result is None:
        return

    if console is None:
        console = get_rich_console()

    if not result.benchmarked:
        if result.benchmarking_error:
            console.print(f"[red]Benchmarking failed:[/red] {result.benchmarking_error}")
        return

    table = create_benchmark_table(result, title)
    if table:
        console.print(table)


def print_success_panel(
    message: str,
    console: Console | None = None,
    title: str = "✓ SUCCESS",
) -> None:
    if console is None:
        console = get_rich_console()
    console.print(
        Panel(
            f"[green]{message}[/green]",
            title=f"[green]{title}[/green]",
            border_style="green",
            box=box.ROUNDED,
        )
    )


def print_error_panel(
    message: str,
    console: Console | None = None,
    title: str = "✗ FAILED",
) -> None:
    if console is None:
        console = get_rich_console()
    console.print(
        Panel(
            f"[red]{message}[/red]",
            title=f"[red]{title}[/red]",
            border_style="red",
            box=box.ROUNDED,
        )
    )


def get_last_error(logs: list[LogMessage] | None) -> LogMessage | None:
    if not logs:
        return None
    for log in reversed(logs):
        if log.type == StepStatus.failed and log.message:
            return log
    return None


def print_validation_result(
    status: ProblemValidationTaskStatus,
    *,
    task_id: str | None = None,
    console: Console | None = None,
    show_benchmark: bool = True,
) -> None:
    """Display validation result with consistent formatting."""
    if console is None:
        console = get_rich_console()

    if task_id:
        console.print(f"[dim]Task:[/dim] {task_id}")
        console.print()

    if status.status == StepStatus.completed:
        print_success_panel(
            f"Validation completed successfully!\n\n[dim]Problem ID:[/dim] [cyan]{status.problem_id}[/cyan]",
        )

        if show_benchmark and status.benchmarking_result:
            console.print()
            print_benchmark_result(status.benchmarking_result)

    elif status.status == StepStatus.failed:
        print_error_panel("Validation failed\n\nCheck the error logs below for details.")

        if status.error_logs:
            console.print()
            print_logs(status.error_logs)

        if status.compilation_result and status.compilation_result.compilation_error:
            console.print()
            console.print("[red]Compilation error:[/red]")
            console.print(textwrap.indent(status.compilation_result.compilation_error, "    "))

        if status.preparation_result and status.preparation_result.preparation_error:
            console.print()
            console.print("[red]Preparation error:[/red]")
            console.print(textwrap.indent(status.preparation_result.preparation_error, "    "))

        if status.benchmarking_result and status.benchmarking_result.benchmarking_error:
            console.print()
            console.print("[red]Benchmarking error:[/red]")
            console.print(textwrap.indent(status.benchmarking_result.benchmarking_error, "    "))

    elif status.status == StepStatus.cancelled:
        console.print("[yellow]Validation was cancelled.[/yellow]")

    else:
        console.print(f"[yellow]Unexpected status: {status.status}[/yellow]")



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from pathlib import Path
from typing import Annotated

import typer

from ..models.openapi import EvaluateKernelRequest, KernelEvaluationDetails, KernelEvaluationStatus, Unit
from ..models.internal import TargetDevice
from ..web.auth import ensure_authenticated, get_current_credentials
from ..web.conn import open_connection


async def cli_evaluate_async(
    reference_file: Path,
    optimized_file: Path,
    device: TargetDevice,
    url: str | None = None,
) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    if not reference_file.exists():
        typer.echo(f"Error: File not found: {reference_file}", err=True)
        raise typer.Exit(1)
    if not optimized_file.exists():
        typer.echo(f"Error: File not found: {optimized_file}", err=True)
        raise typer.Exit(1)

    hardware_provider, hardware_model = device.to_api_device().split(":")

    request = EvaluateKernelRequest(
        reference_code=reference_file.read_text(),
        optimized_code=optimized_file.read_text(),
        name=optimized_file.name,
        origin="user",
        extras={},
    )

    typer.echo("Evaluating code...")

    try:
        async with open_connection(url) as conn:
            await ensure_authenticated(conn)
            response = await conn.post(
                f"kernel-evaluation/evaluation/{hardware_provider}/{hardware_model}",
                request,
                reply_format=KernelEvaluationDetails,
                token=creds.token,
            )
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1) from e

    evaluation = response.evaluation
    if evaluation is None or evaluation.status != KernelEvaluationStatus.COMPLETED:
        typer.echo("\n✗ Evaluation failed!", err=True)
        raise typer.Exit(1)

    optimized_time = evaluation.optimized_time
    reference_time = evaluation.reference_time
    speedup = evaluation.speedup

    # Default to milliseconds when unit is absent.
    unit = evaluation.optimized_time_unit or evaluation.reference_time_unit or Unit.ms
    unit_value = unit.value

    typer.echo("\n✓ Evaluation successful!")
    typer.echo("\nBenchmark Results:")
    if reference_time is not None:
        typer.echo(f"  Reference time: {reference_time:.6f} {unit_value}")
    if optimized_time is not None:
        typer.echo(f"  Solution time:  {optimized_time:.6f} {unit_value}")
    if speedup is not None:
        typer.echo(f"  Speedup:        {speedup:.2f}x")


def cli_evaluate(
    reference_file: Annotated[Path, typer.Argument(help="Path to file containing reference code.")],
    optimized_file: Annotated[Path, typer.Argument(help="Path to file containing optimized code.")],
    device: Annotated[TargetDevice, typer.Option("-d", "--device", help="Device type.")],
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """Evaluate code against a reference implementation on remote hardware."""
    asyncio.run(
        cli_evaluate_async(reference_file=reference_file, optimized_file=optimized_file, device=device, url=url)
    )



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime, timezone

from rich.table import Table
from rich import box

from ..models.openapi import StepStatus, KernelEvaluationStatus, Unit
from ..models.internal import TargetDevice


STATUS_COLORS = {
    "in_progress": "yellow",
    "completed": "green",
    "not_started": "bright_black",
    "failed": "red",
    "cancelled": "red",
}

STATUS_INDICATORS = {
    "in_progress": "*",
    "completed": "+",
    "not_started": "o",
    "failed": "x",
    "cancelled": "x",
}

STATUS_LABELS = {
    "in_progress": "running",
    "completed": "done",
    "not_started": "queued",
    "failed": "failed",
    "cancelled": "cancelled",
}


DEVICE_LABELS = {
    TargetDevice.H100: "NVIDIA H100",
    TargetDevice.H200: "NVIDIA H200",
    TargetDevice.B200: "NVIDIA B200",
    TargetDevice.L40S: "NVIDIA L40S",
    TargetDevice.MI300X: "AMD MI300X",
    TargetDevice.ADRENO_750: "Snapdragon 8 Gen 3 - GPU",
    TargetDevice.ADRENO_830: "Snapdragon 8 Elite - GPU",
    TargetDevice.HEXAGON_V75: "Snapdragon 8 Gen 3 - NPU",
    TargetDevice.HEXAGON_V79: "Snapdragon 8 Elite - NPU",
}


def format_status(status: StepStatus | KernelEvaluationStatus | str | None) -> str:
    if isinstance(status, (StepStatus, KernelEvaluationStatus)):
        status_str = status.value.lower()
    else:
        status_str = status if status else "[dim]unknown[/dim]"

    color = STATUS_COLORS.get(status_str, "white")
    indicator = STATUS_INDICATORS.get(status_str, "?")
    label = STATUS_LABELS.get(status_str, status_str)
    return f"[{color}]{indicator} {label}[/{color}]"


def format_close_miss_status(best_atol: float | None, best_rtol: float | None) -> str:
    tol_str = f" (a:{best_atol} r:{best_rtol})" if best_atol is not None else ""
    return f"[yellow]~ low precision{tol_str}[/yellow]"


def format_time_ago(dt: datetime | None) -> str:
    if dt is None:
        return "-"
    now = datetime.now(timezone.utc)
    seconds = (now - dt).total_seconds()
    if seconds < 0:
        return "now"
    if seconds < 60:
        return f"{int(seconds)}s ago"
    elif seconds < 3600:
        return f"{int(seconds / 60)}m ago"
    elif seconds < 86400:
        return f"{int(seconds / 3600)}h ago"
    else:
        return f"{int(seconds / 86400)}d ago"


def format_device(device: TargetDevice | str | None) -> str:
    if isinstance(device, TargetDevice):
        return DEVICE_LABELS.get(device, "[dim]unknown[/dim]")

    return device or "-"


def format_time(value: float | None, unit: Unit | None = None) -> str:
    if value is None:
        return "-"
    label = unit.value if unit else "ms"
    return f"{value:.3f} {label}"


def format_speedup(ratio: float | None) -> str:
    if ratio is None:
        return "-"
    if ratio >= 1.0:
        return f"[green]{ratio:.2f}x faster[/green]"
    else:
        slowdown = 1.0 / ratio
        if ratio >= 0.5:
            return f"[yellow]{slowdown:.2f}x slower[/yellow]"
        else:
            return f"[red]{slowdown:.2f}x slower[/red]"


def create_styled_table(title: str | None = None) -> Table:
    """Create a table with consistent styling."""
    return Table(
        title=title,
        box=box.SIMPLE,
        header_style="bold cyan",
        border_style="dim",
        title_style="bold white",
        show_lines=False,
        padding=(0, 1),
    )



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import textwrap
import itertools as itr
from pathlib import Path
from typing import Annotated
from uuid import UUID

import typer
from rich.syntax import Syntax
from rich.table import Table

from ..utils import get_rich_console
from ..web.conn import open_connection
from ..web.auth import get_current_credentials
from ..web.sessions import get_user_sessions, get_session_kernels, resolve_session
from ..models.openapi import EvaluatedKernel
from ..components.strings import (
    create_styled_table,
    format_close_miss_status,
    format_status,
    format_time,
    format_speedup,
)


def create_kernels_table(kernels: list[list[EvaluatedKernel]], title: str = "Kernels") -> Table:
    table = create_styled_table(title)

    table.add_column("Attempt", justify="center")
    table.add_column("Kernel ID", style="cyan", no_wrap=True)
    table.add_column("Name", no_wrap=True, max_width=15)
    table.add_column("Status", no_wrap=True)
    table.add_column("Time", justify="right")
    table.add_column("vs torch.compile", justify="right")

    for attempt_num, attempt_kernels in enumerate(kernels):
        for kernel in attempt_kernels:
            kernel_id = str(kernel.id)[:8]
            name = kernel.name
            if len(name) > 15:
                name = name[:12] + "..."
            speedup = kernel.speed_up_compiled

            if kernel.is_close_miss:
                status_display = format_close_miss_status(kernel.best_atol, kernel.best_rtol)
            else:
                status_display = format_status(kernel.evaluation_status)

            table.add_row(
                str(attempt_num + 1),
                kernel_id,
                name,
                status_display,
                format_time(kernel.time, kernel.time_unit),
                format_speedup(speedup),
            )

    return table


async def resolve_kernel(
    kernels: list[list[EvaluatedKernel]],
    session_id: UUID,
    kernel_id: str,
) -> EvaluatedKernel | None:
    """Find kernel matching the given ID prefix. Returns (full_kernel_id, kernels_data)."""

    matches: list[EvaluatedKernel] = []
    for krn in itr.chain.from_iterable(kernels):
        if str(krn.id).startswith(kernel_id):
            matches.append(krn)

    if not matches:
        return None
    elif len(matches) > 1:
        kernel_list = [" * " + str(k.id) for k in matches]
        kernel_block = textwrap.indent("\n".join(kernel_list), "    ")
        raise ValueError(
            f"Kernel ID prefix: {kernel_id} is matching more than one kernel from session {session_id!r}:\n"
            + kernel_block
        )

    return matches[0]


async def cli_kernels_list_async(session_id: str, url: str | None = None) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    console = get_rich_console()

    async with open_connection(url) as conn:
        sessions = await get_user_sessions(conn)
        match = await resolve_session(sessions, session_id)

        if not match:
            raise SystemExit(f"No session found matching '{session_id}'")

        kernels = await get_session_kernels(conn, str(match.id))

    if not kernels:
        console.print("[dim]No kernels found for this session.[/dim]")
        return

    table = create_kernels_table(kernels, title=f"Kernels for {str(match.id)[:8]} ({match.label})")
    console.print(table)


async def cli_kernels_code_async(session_id: str, kernel_id: str, output: str | None, url: str | None = None) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    console = get_rich_console()

    async with open_connection(url) as conn:
        sessions = await get_user_sessions(conn)
        match = await resolve_session(sessions, session_id)

        if not match:
            raise SystemExit(f"No session found matching '{session_id}'")

        kernels = await get_session_kernels(conn, str(match.id))

        # Resolve kernel ID and get perf data from list endpoint
        kernel = await resolve_kernel(kernels, match.id, kernel_id)

        if not kernel:
            console.print(f"[red]No kernel found matching '{kernel_id}'[/red]")
            if kernels:
                console.print("\nAvailable kernels:")
                for krn in itr.chain.from_iterable(kernels):
                    console.print(f"  {str(krn.id)[:8]}  {krn.name}")

            raise SystemExit(1)

    if not kernel.code:
        raise SystemExit("No code available for this kernel.")

    # Save to file if requested
    if output:
        Path(output).write_text(kernel.code, encoding="utf-8")
        console.print(f"[green]Kernel saved to: {output}[/green]")
        return

    # Display with syntax highlighting (no box for easy copy/paste)
    lexer = "python"

    syntax = Syntax(
        kernel.code,
        lexer,
        theme="monokai",
        line_numbers=False,
        word_wrap=False,
    )
    console.print(syntax)

    # Show performance data
    console.print()
    console.print(f"[bold cyan]── {kernel.name or 'Kernel'} ({str(kernel.id)[:8]}) ──[/bold cyan]")

    if kernel.is_close_miss:
        console.print(f"  Status:           {format_close_miss_status(kernel.best_atol, kernel.best_rtol)}")

    time_ms = kernel.time
    ref_eager = kernel.reference_eager
    ref_compiled = kernel.reference_compile
    speedup_eager = kernel.speed_up_eager
    speedup_compiled = kernel.speed_up_compiled

    if time_ms:
        console.print(f"  Kernel time:      [cyan]{time_ms:.3f} ms[/cyan]")
    if ref_eager:
        console.print(f"  Reference eager:  [dim]{ref_eager:.3f} ms[/dim]")
    if ref_compiled:
        console.print(f"  torch.compile:    [dim]{ref_compiled:.3f} ms[/dim]")
    if speedup_eager:
        console.print(f"  vs eager:         {format_speedup(speedup_eager)}")
    if speedup_compiled:
        console.print(f"  vs torch.compile: {format_speedup(speedup_compiled)}")


def cli_kernels(
    session_id: Annotated[str, typer.Argument(help="Session ID (or prefix).")],
    kernel_id: Annotated[
        str | None,
        typer.Argument(help="Kernel ID (or prefix) - if provided, shows kernel code."),
    ] = None,
    output: Annotated[str | None, typer.Option("-o", "--output", help="Save kernel code to file.")] = None,
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """List kernels for a session, or view kernel code if kernel_id is provided."""
    if kernel_id:
        asyncio.run(cli_kernels_code_async(session_id, kernel_id, output, url))
    else:
        asyncio.run(cli_kernels_list_async(session_id, url))



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..utils import get_env_vars, get_rich_console
from ..components.logo import print_header
from ..components.strings import create_styled_table
from ..web.auth import get_current_credentials


def cli_info() -> None:
    """Shows current information about user status and env variables."""
    from ..version import info

    console = get_rich_console()

    print_header(console)

    ver_info = info()

    console.print("[bold]Makora version[/bold]:", ver_info["version"])
    if ver_info["repo"] != "unknown":
        console.print("[dim]    Repo: " + ver_info["repo"] + "[/dim]")
        console.print("[dim]    Commit: " + ver_info["commit"] + "[/dim]")
    console.print()

    user = get_current_credentials()
    if user is not None:
        console.print("[bold]Logged in as:[/bold]", f"[cyan]{user.user}[/cyan]")
    else:
        console.print("[bold]Currently logged out[/bold]")

    console.print()
    console.print("[bold]Existing knobs and their values (non-value in red):[/bold]")

    table = create_styled_table()
    table.add_column("Env Variable")
    table.add_column("Value")
    table.add_column("Default")

    for var in get_env_vars():
        if var.hidden:
            continue

        not_default = var.value != var.default
        if not_default:
            value = f"[red bold]{var.value}[/red bold]"
        else:
            value = var.value

        table.add_row(var.var_name, value, f"[dim]{var.default}[/dim]")

    console.print(table)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from typing import Annotated
from uuid import UUID

import typer
from rich.table import Table
from rich import box

from ..utils import get_rich_console
from ..web.conn import open_connection
from ..web.auth import get_current_credentials
from ..web.sessions import fetch_session_extra, get_user_sessions, stop_job, resolve_session
from ..models.openapi import AgentSessionSummary
from ..models.internal import SessionExtra
from ..components.strings import format_status, format_time_ago, format_device, format_speedup


def create_jobs_table(
    sessions: list[AgentSessionSummary],
    extras: dict[UUID, SessionExtra] | None = None,
    title: str = "Jobs",
) -> Table:
    table = Table(
        title=title,
        box=box.SIMPLE,
        header_style="bold cyan",
        border_style="dim",
        title_style="bold white",
        show_lines=False,
        padding=(0, 1),
    )

    table.add_column("Session ID", style="cyan", no_wrap=True)
    table.add_column("Status", no_wrap=True)
    table.add_column("Label", no_wrap=True, max_width=20)
    if extras is not None:
        table.add_column("Device", no_wrap=True)
        table.add_column("vs torch.compile", justify="right")
    table.add_column("Started", justify="right")

    for session in sessions:
        session_id = str(session.id)[:8]
        status_display = format_status(session.status)
        label = session.label or "-"
        if len(label) > 20:
            label = label[:17] + "..."
        started = format_time_ago(session.started_at)

        row = [session_id, status_display, label]
        if extras is not None:
            extra = extras.get(session.id, SessionExtra())
            row.append(format_device(extra.device))
            row.append(format_speedup(extra.speedup))
        row.append(started)

        table.add_row(*row)

    return table


async def cli_stop_async(job_uuid: str, url: str | None = None) -> None:
    if get_current_credentials() is None:
        raise SystemExit("You need to login first with 'makora login'")

    console = get_rich_console()

    async with open_connection(url) as conn:
        sessions = await get_user_sessions(conn)
        session = await resolve_session(sessions, job_uuid)
        if session is None:
            console.print(f"[red]Job {job_uuid} not found.[/red]")
            raise SystemExit(1)
        session_id = session.id
        console.print(f"Found job: [cyan]{session_id}[/cyan]")
        if await stop_job(conn, session_id):
            console.print(f"[green]Job {str(session_id)[:8]} stopped successfully.[/green]")
        else:
            console.print(f"[yellow]Job {str(session_id)[:8]} is not running.[/yellow]")


async def cli_jobs_async(fast: bool, url: str | None = None) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    console = get_rich_console()
    async with open_connection(url) as conn:
        sessions = await get_user_sessions(conn)

        if not sessions:
            console.print("[dim]No jobs found.[/dim]")
            return

        extras: dict[UUID, SessionExtra] | None = None
        if not fast:
            # Fetch extras in parallel
            tasks = [fetch_session_extra(conn, s.id) for s in sessions]
            results = await asyncio.gather(*tasks)
            extras = {s.id: r for s, r in zip(sessions, results) if r is not None}

    table = create_jobs_table(sessions, extras)
    console.print(table)


def cli_jobs(
    fast: Annotated[bool, typer.Option(help="Skip fetching extra data (device, speedup).")] = False,
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """Lists jobs created by the user."""
    asyncio.run(cli_jobs_async(fast, url))


def cli_stop(
    job_uuid: Annotated[str, typer.Argument(help="The UUID (or prefix) of the job to stop.")],
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """Request to stop a running job."""
    asyncio.run(cli_stop_async(job_uuid, url))



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from pathlib import Path
from typing import Annotated

import typer

from ..models.openapi import (
    KernelEvaluationStatus,
    KernelProfilingDetails,
    ProfileKernelRequest,
    ProfilingMode,
)
from ..models.internal import TargetDevice
from ..web.auth import ensure_authenticated, get_current_credentials
from ..web.conn import open_connection


def _extract_error(details: KernelProfilingDetails) -> str:
    run = details.kernel_profiling_run
    if run is None:
        return "No profiling details returned by server."
    if run.profiling_result and run.profiling_result.error:
        return run.profiling_result.error.message
    if run.compilation_result and run.compilation_result.error:
        return run.compilation_result.error.message
    if run.preparation_result and run.preparation_result.error:
        return run.preparation_result.error.message
    return "Profiling did not complete successfully."


async def cli_profile_async(
    reference_file: Path, optimized_file: Path, device: TargetDevice, url: str | None = None
) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    if not reference_file.exists():
        typer.echo(f"Error: File not found: {reference_file}", err=True)
        raise typer.Exit(1)

    if not optimized_file.exists():
        typer.echo(f"Error: File not found: {optimized_file}", err=True)
        raise typer.Exit(1)

    hardware_provider, hardware_model = device.to_api_device().split(":")

    request = ProfileKernelRequest(
        reference_code=reference_file.read_text(),
        optimized_code=optimized_file.read_text(),
        name=optimized_file.name,
        origin="user",
        extras={},
        mode=ProfilingMode.full,
    )

    typer.echo("Profiling code...")

    try:
        async with open_connection(url) as conn:
            await ensure_authenticated(conn)
            response = await conn.post(
                f"kernel-evaluation/profile/{hardware_provider}/{hardware_model}",
                request,
                reply_format=KernelProfilingDetails,
                token=creds.token,
            )
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1) from e

    run = response.kernel_profiling_run
    if run is None or run.status != KernelEvaluationStatus.COMPLETED:
        typer.echo("\nProfiling failed!", err=True)
        typer.echo(f"\nError: {_extract_error(response)}", err=True)
        raise typer.Exit(1)

    profiling_result = run.profiling_result
    if profiling_result is None:
        typer.echo("\nNo kernel profiling data available.")
        return

    typer.echo("\nProfiling successful!")
    kernels = profiling_result.kernel_info or []
    if not kernels:
        typer.echo("\nNo kernel profiling data available.")
        return

    typer.echo(f"\nProfiled {len(kernels)} kernel(s):\n")
    for i, kernel in enumerate(kernels, 1):
        typer.echo(f"--- Kernel {i} ---")
        if kernel.raw_metrics:
            typer.echo("\nMetrics:")
            for key, value in kernel.raw_metrics.items():
                typer.echo(f"  {key}: {value}")
        if kernel.details_page_text:
            typer.echo("\nDetails:")
            typer.echo(kernel.details_page_text)
        if kernel.nsys_report_text:
            typer.echo("\nNsys Report:")
            typer.echo(kernel.nsys_report_text)
        typer.echo()


def cli_profile(
    reference_file: Annotated[Path, typer.Argument(help="Path to file containing reference code.")],
    optimized_file: Annotated[Path, typer.Argument(help="Path to file containing optimized code.")],
    device: Annotated[TargetDevice, typer.Option("-d", "--device", help="Device type.")],
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """Profile code using the remote Makora evaluator."""
    asyncio.run(cli_profile_async(reference_file=reference_file, optimized_file=optimized_file, device=device, url=url))



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import asyncio
import getpass
from typing import Annotated
from pathlib import Path

import typer

from ..utils import get_rich_console
from ..config import GENERATE_BASE_URL
from ..web.conn import open_connection
from ..web.auth import login_with_token


async def cli_login_async(
    user: str | None,
    token: str | None = None,
    interactive: bool = True,
    url: str | None = None,
) -> None:
    console = get_rich_console()

    if token is None:
        if not interactive:
            console.print(
                "[red]You have to provide your user's token, or a "
                "file containing it, using --token when running "
                "non-interactively.[/red]"
            )
            raise typer.Exit(1)

        console.print(
            f"Please go to: {url or GENERATE_BASE_URL.value}/tokens and select an "
            "API token you want to use for this session."
        )
        console.print("Create a new token if none shows up. You need to have an account and login to access the page.")
        console.print()
        token = getpass.getpass("Token: ")

    maybe_token_file = Path(token)
    if maybe_token_file.exists():
        if not maybe_token_file.is_file():
            console.print(f"[red]Path {token!r} exists but is not a file! Cannot read the user's token.[/red]")
            raise typer.Exit(1)

        token = maybe_token_file.read_text()
        if not token:
            console.print(f"[red]Token file at: {str(maybe_token_file)!r} is empty.[/red]")
            raise typer.Exit(1)

    async with open_connection(url) as conn:
        creds = await login_with_token(conn, user, token)
        if creds.token == token:
            console.print("[green][bold]Logged in successfully![/]")
        else:
            console.print(
                "[red]Warning: Login succeeded but resulted in a different token "
                "being in-use than the provided one! This is unexpected and might "
                "result in further errors.[/red]"
            )


def cli_login(
    user: Annotated[
        str | None,
        typer.Option(
            help="Your user's email address. Optional: if provided will make "
            "sure the user associated with the provided token matches the given user."
        ),
    ] = None,
    token: Annotated[
        str | None,
        typer.Option(
            help="Your user's API token. If not provided, the app will ask for "
            "it interactively (or fail if running non-interactively). If "
            "provided, can be the token itself or a path to a file containing "
            "the token."
        ),
    ] = None,
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
    alone: Annotated[bool, typer.Option(help="Disables any interactivity.")] = False,
) -> None:
    """Login command to the service."""
    asyncio.run(
        cli_login_async(
            user=user,
            token=token,
            interactive=(not alone and sys.stdin.isatty()),
            url=url,
        )
    )



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from types import TracebackType
from typing import TypeVar, Any
from typing_extensions import Self

import aiohttp
from pydantic import BaseModel

from ..config import get_generate_base_url
from .errors import map_errors


T = TypeVar("T", bound=BaseModel)


class Connection:
    def __init__(self, base_url: str) -> None:
        if not base_url:
            raise ValueError("empty base_url")
        if not base_url.endswith("/"):
            base_url += "/"

        self.base_url = base_url
        self.client: aiohttp.ClientSession | None = None

    async def __aenter__(self) -> Self:
        client = aiohttp.ClientSession(base_url=self.base_url, raise_for_status=False)
        await client.__aenter__()
        self.client = client
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self.client is not None
        await self.client.__aexit__(exc_type, exc_val, exc_tb)
        self.client = None
        return

    async def _reconnect(self) -> None:
        if self.client is not None:
            await self.client.__aexit__(None, None, None)
            self.client = None
        client = aiohttp.ClientSession(base_url=self.base_url, raise_for_status=False)
        await client.__aenter__()
        self.client = client

    async def post(
        self,
        endpoint: str,
        *payload: BaseModel,
        reply_format: type[T],
        token: str | None = None,
        json: bool = True,
    ) -> T:
        if self.client is None:
            raise ValueError("Connection has not been opened or has been already closed")

        kwargs = {}
        if token is not None:
            kwargs["headers"] = {"Authorization": f"Bearer {token}"}

        if payload:
            params: dict[str, Any] = {}
            for p in payload:
                params.update(p.model_dump(mode="json"))

            if json:
                kwargs["json"] = params
            else:
                kwargs["data"] = params

        try:
            async with self.client.post(
                endpoint,
                **kwargs,  # type: ignore[arg-type,unused-ignore]
            ) as resp:
                await map_errors(resp)
                repl = await resp.text()
                return reply_format.model_validate_json(repl)
        except aiohttp.ServerDisconnectedError:
            await self._reconnect()
            async with self.client.post(
                endpoint,
                **kwargs,  # type: ignore[arg-type,unused-ignore]
            ) as resp:
                await map_errors(resp)
                repl = await resp.text()
                return reply_format.model_validate_json(repl)

    async def get(self, endpoint: str, reply_format: type[T], token: str | None = None) -> T:
        if self.client is None:
            raise ValueError("Connection has not been opened or has been already closed")

        headers: dict[str, str] = {}
        if token is not None:
            headers["Authorization"] = f"Bearer {token}"

        try:
            async with self.client.get(endpoint, headers=headers) as resp:
                await map_errors(resp)
                repl = await resp.text()
                return reply_format.model_validate_json(repl)
        except aiohttp.ServerDisconnectedError:
            await self._reconnect()
            async with self.client.get(endpoint, headers=headers) as resp:
                await map_errors(resp)
                repl = await resp.text()
                return reply_format.model_validate_json(repl)


def open_connection(url: str | None = None) -> Connection:
    return Connection(get_generate_base_url(url))



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import sys
from pathlib import Path
from typing import Annotated

import typer
from rich.syntax import Syntax

from ..utils import get_rich_console
from ..models.internal import TargetDevice
from ..web.conn import open_connection
from ..web.auth import ensure_authenticated
from ..components.logo import print_mini_header
from ..components.problem_validation import validate_problem


async def cli_check_async(
    file: Path,
    device: TargetDevice,
    url: str | None = None,
    fix: bool = False,
    interactive: bool = True,
) -> None:
    console = get_rich_console()
    print_mini_header(f"Check: {file.name}")

    if not file.exists():
        console.print(f"[red]Error:[/red] File not found: {file}")
        raise typer.Exit(1)

    code = file.read_text()

    async with open_connection(url) as conn:
        await ensure_authenticated(conn)

        hint_command = f"makora check --file {file} --device {device.value}"
        validation_result = await validate_problem(
            conn=conn,
            code=code,
            device=device,
            label=file.name,
            fix=fix,
            interactive=interactive,
            hint_command=hint_command,
            console=console,
        )

        if validation_result is None:
            raise typer.Exit(1)

        _, final_code, fixed = validation_result

        if fixed:
            # Display with syntax highlighting (no box for easy copy/paste)
            lexer = "python"
            syntax = Syntax(
                final_code,
                lexer,
                theme="monokai",
                line_numbers=False,
                word_wrap=False,
            )
            console.print()
            console.print("[bold]Final working solution, after applying fixes:[/bold]")
            console.print(syntax)


def cli_check(
    file: Annotated[Path, typer.Argument(help="Path to Python file to validate.")],
    device: Annotated[TargetDevice, typer.Option("-d", "--device", help="Device type.")],
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
    fix: Annotated[
        bool,
        typer.Option(
            help="Enables automatic fix suggestions if the provided problem file "
            "contains errors. Any fix has to be accepted by the user interactively, "
            "unless --alone if passed or input is not a TTY, in which case all fixes "
            "will be accepted automatically."
        ),
    ] = False,
    alone: Annotated[bool, typer.Option(help="Disables any interactivity.")] = False,
) -> None:
    """Evaluates the given reference catching possible errors."""
    asyncio.run(
        cli_check_async(
            file=file,
            device=device,
            url=url,
            fix=fix,
            interactive=(not alone and sys.stdin.isatty()),
        )
    )



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any

from aiohttp import ClientResponse


class HttpError(ValueError):
    def __init__(self, code: int, url: str, *args: Any) -> None:
        super().__init__(code, url, *args)
        self.code = code
        self.url = url


class Http404(HttpError):
    def __init__(self, url: str, *args: Any) -> None:
        super().__init__(404, url, *args)


async def map_errors(resp: ClientResponse) -> None:
    if resp.status in {200, 201}:
        return

    if resp.content_type == "application/json":
        data = await resp.json()
    else:
        data = await resp.text()

    match resp.status:
        case 404:
            raise Http404(str(resp.real_url), data)
        case _:
            raise HttpError(resp.status, str(resp.real_url), data)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from pathlib import Path
from typing import Annotated

import typer
from rich.syntax import Syntax

from ..utils import get_rich_console
from ..web.conn import open_connection
from ..web.auth import get_current_credentials
from ..web.sessions import get_user_sessions, get_session, resolve_session


async def cli_refcode_async(session_id: str, output: str | None, url: str | None) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    console = get_rich_console()

    async with open_connection(url) as conn:
        sessions = await get_user_sessions(conn)
        match = await resolve_session(sessions, session_id)

        if not match:
            raise SystemExit(f"No session found matching '{session_id}'")

        session = await get_session(conn, str(match.id))

    code = session.request.problem_description_code.code
    if not code:
        raise SystemExit("No refcode available for this session.")

    if output:
        Path(output).write_text(code, encoding="utf-8")
        console.print(f"[green]Refcode saved to: {output}[/green]")
        return

    syntax = Syntax(
        code,
        "python",
        theme="monokai",
        line_numbers=False,
        word_wrap=False,
    )
    console.print(syntax)


def cli_refcode(
    session_id: Annotated[str, typer.Argument(help="Session ID (or prefix).")],
    output: Annotated[str | None, typer.Option("-o", "--output", help="Save refcode to a file.")] = None,
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
) -> None:
    """Show the original refcode submitted for a session."""
    asyncio.run(cli_refcode_async(session_id, output, url))



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration management for Makora CLI."""

from .utils import EnvVar


GENERATE_BASE_URL = EnvVar("MAKORA_URL", "https://generate.stage.makora.com")


def _normalize_generate_api_url(url: str) -> str:
    normalized = url.strip().rstrip("/")
    if normalized.endswith("/api/v1"):
        return f"{normalized}/"
    return f"{normalized}/api/v1/"


def get_generate_base_url(url: str | None = None) -> str:
    """Get normalized base URL for Generate API requests."""
    return _normalize_generate_api_url(url or GENERATE_BASE_URL.value)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import sys
from pathlib import Path
from typing import Annotated

import typer

from ..utils import get_rich_console
from ..models.openapi import KernelLanguage
from ..models.internal import TargetDevice
from ..web.conn import open_connection
from ..web.auth import ensure_authenticated
from ..web.sessions import new_session
from ..components.spinner import show_spinner
from ..components.logo import print_mini_header
from ..components.results import print_success_panel
from ..components.problem_validation import validate_problem


async def cli_submit_async(
    file: Path,
    device: TargetDevice,
    language: KernelLanguage | None,
    label: str,
    atol: float,
    rtol: float,
    url: str | None,
    fix: bool = False,
    instr: list[Path] | None = None,
    interactive: bool = True,
) -> None:
    console = get_rich_console()
    print_mini_header(f"Submit: {file.name}")

    if not file.exists():
        console.print(f"[red]Error:[/red] File not found: {file}")
        raise typer.Exit(1)

    code = file.read_text()

    user_prompt = ""
    if instr:
        for instr_file in instr:
            if not instr_file.exists():
                console.print(f"[red]Provided instruction file: {instr_file} could not be found")
                raise typer.Exit(1)

            user_prompt += instr_file.read_text().strip() + "\n\n"

        user_prompt = user_prompt.strip()

    if language is None:
        language = device.get_default_language()

    elif not device.supports_language(language):
        console.print(f"[red]Error:[/red] Device {device.value} does not support {language.value}")
        raise typer.Exit(1)

    console.print(f"[dim]Device:[/dim] {device.value}")
    console.print(f"[dim]Language:[/dim] {language.value}")
    console.print()

    async with open_connection(url) as conn:
        await ensure_authenticated(conn)

        hint_command = f"makora submit --file {file} --device {device.value}"
        validation_result = await validate_problem(
            conn=conn,
            code=code,
            device=device,
            label=label,
            fix=fix,
            interactive=interactive,
            hint_command=hint_command,
            console=console,
        )

        if validation_result is None:
            raise typer.Exit(1)

        problem_id, *_ = validation_result

        console.print()
        with show_spinner("Creating session..."):
            session_id = await new_session(
                conn=conn,
                problem_id=problem_id,
                language=language,
                device=device,
                label=label,
                atol=atol,
                rtol=rtol,
                user_prompt=user_prompt,
            )

    console.print()
    print_success_panel(
        f"Session created!\n\n"
        f"[dim]Session ID:[/dim] [cyan]{session_id}[/cyan]\n"
        f"[dim]Problem ID:[/dim] [cyan]{problem_id}[/cyan]"
    )
    console.print()
    console.print("[dim]Monitor progress with:[/dim] [cyan]makora jobs[/cyan]")


def cli_submit(
    file: Annotated[Path, typer.Option(help="A file containing problem definition (reference code).")],
    device: Annotated[TargetDevice, typer.Option("-d", "--device", help="Device type.")],
    language: Annotated[
        KernelLanguage | None,
        typer.Option(help="Target language. If not provided will use the default for the given device."),
    ] = None,
    label: Annotated[str, typer.Option(help="Optional user-defined label for the job.")] = "",
    atol: Annotated[float, typer.Option(help="Absolute tolerance to use when validating generated solutions.")] = 1e-2,
    rtol: Annotated[float, typer.Option(help="Relative tolerance to use when validating generated solutions.")] = 1e-2,
    url: Annotated[
        str | None,
        typer.Option(
            help="Overwrite the base URL used to communicate with the service. If "
            "not provided will use the one controlled by MAKORA_URL env var. "
            "Use `makora info` for its value."
        ),
    ] = None,
    fix: Annotated[
        bool,
        typer.Option(
            help="Enables automatic fix suggestions if the provided problem file "
            "contains errors. Any fix has to be accepted by the user interactively, "
            "unless --alone if passed or input is not a TTY, in which case all fixes "
            "will be accepted automatically."
        ),
    ] = False,
    instr: Annotated[
        list[Path] | None,
        typer.Option(
            help="A path to a markdown file containing additional instructions that will "
            "be passed to the agent. Can be specified multiple times, in which case "
            "all files' content will be concatenated, in the order in which the files "
            "are specified."
        ),
    ] = None,
    alone: Annotated[bool, typer.Option(help="Disables any interactivity.")] = False,
) -> None:
    """Submit a anew job to the agent."""
    asyncio.run(
        cli_submit_async(
            file=file,
            device=device,
            language=language,
            label=label,
            atol=atol,
            rtol=rtol,
            url=url,
            fix=fix,
            instr=instr,
            interactive=(not alone and sys.stdin.isatty()),
        )
    )



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .login import cli_login
from .logout import cli_logout
from .info import cli_info
from .submit import cli_submit
from .jobs import cli_jobs, cli_stop
from .kernels import cli_kernels
from .check import cli_check
from .evaluate import cli_evaluate
from .refcode import cli_refcode
from .install import cli_install
from .profile import cli_profile
from .expert_generate import cli_expert_generate
from .document_search import cli_document_search


__all__ = [
    "cli_login",
    "cli_logout",
    "cli_info",
    "cli_submit",
    "cli_jobs",
    "cli_stop",
    "cli_kernels",
    "cli_check",
    "cli_refcode",
    "cli_profile",
    "cli_evaluate",
    "cli_expert_generate",
    "cli_document_search",
    "cli_install",
]



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import UUID

from rich.console import Console

from ..utils import get_rich_console
from ..web.conn import Connection
from ..web.problems import submit_and_poll_validation
from ..models.internal import TargetDevice
from ..models.openapi import StepStatus
from ..components.spinner import show_spinner
from ..components.fix_suggestions import print_fix_suggestion, prompt_accept_fix
from ..components.results import print_validation_result


STEP_TO_SPINNER_LABEL = {
    "compilation": "Compiling...",
    "preparation": "Preparing objects for execution...",
    "benchmarking": "Benchmarking...",
    "Suggesting fixes": "Suggesting fixes...",
}


async def validate_problem(
    conn: Connection,
    code: str,
    device: TargetDevice,
    label: str,
    fix: bool,
    interactive: bool,
    hint_command: str | None = None,
    console: Console | None = None,
) -> tuple[UUID, str, bool] | None:
    if console is None:
        console = get_rich_console()

    revalidating = False
    while True:
        spinner_message = "Re-testing fixed code..." if revalidating else "Validating problem..."
        with show_spinner(spinner_message) as spinner:

            def on_progress(step: str) -> None:
                if spinner is None:
                    return
                label = STEP_TO_SPINNER_LABEL.get(step, f"{step}...")
                spinner.update(f"[cyan]{label}[/cyan]")

            status = await submit_and_poll_validation(
                conn,
                code,
                label,
                target_device=device,
                fix=fix,
                on_progress=on_progress,
            )

        print_validation_result(status, show_benchmark=True)

        if status.status == StepStatus.completed:
            break

        if not fix:
            if hint_command:
                console.print()
                console.print("[dim]Hint: try submitting with --fix to get automatic fix suggestions:[/dim]")
                console.print(f"[dim]  {hint_command} --fix[/dim]")
            return None

        if status.fix_suggestions is None:
            console.print("[red]No fixes to suggest! Sorry :(")
            return None

        print_fix_suggestion(status.fix_suggestions, code, console)

        if not interactive or prompt_accept_fix(console):
            code = status.fix_suggestions.formatted_code
            revalidating = True
            console.print()
            if not interactive:
                console.print("[dim]Auto-accepting the fix in a non-interactive mode[/dim]")
            console.print("[dim]Resubmitting with fixed code...[/dim]")
            console.print()
        else:
            return None

    if status.problem_id is None:
        console.print("[red]Error:[/red] No problem ID returned")
        return None

    return status.problem_id, code, revalidating



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
from typing import Annotated

import typer

from ..plugins.claude import ClaudeInstallError, install_claude_code_plugin
from ..web.auth import get_current_credentials


class InstallTarget(Enum):
    claude = "claude"


def cli_install(
    target: Annotated[InstallTarget, typer.Argument(help="Platform to install for (currently only 'claude').")],
) -> None:
    """Install the Makora plugin for a supported platform."""
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    if target != InstallTarget.claude:
        typer.echo(f"Unsupported install target: {target}", err=True)
        raise typer.Exit(1)

    typer.echo("Installing Makora plugin for Claude Code...")
    try:
        messages = install_claude_code_plugin()
        if "cache_removed" in messages:
            typer.echo(messages["cache_removed"])
        if "removed" in messages:
            typer.echo(messages["removed"])
        typer.echo(messages["marketplace"])
        typer.echo("Installing makora-plugin...")
        typer.echo(messages["plugin"])
        typer.echo("\nMakora plugin installed successfully for Claude Code!")
    except ClaudeInstallError as e:
        typer.echo(str(e), err=True)
        raise typer.Exit(1)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import difflib

import typer
from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax

from ..utils import get_rich_console
from ..models.openapi import FixSuggestion


def print_fix_suggestion(
    fix_suggestion: FixSuggestion,
    original_code: str | None = None,
    console: Console | None = None,
    language: str = "python",
    theme: str = "monokai",
    no_diffs: bool = False,
) -> None:
    if console is None:
        console = get_rich_console()

    console.print()
    console.print("[yellow]Fix Suggestion:[/yellow]")
    console.print(f"{fix_suggestion.summary}")
    console.print()

    if original_code and not no_diffs:
        to_display = "\n".join(
            difflib.unified_diff(
                original_code.splitlines(),
                fix_suggestion.formatted_code.splitlines(),
                fromfile="Original",
                tofile="New",
                lineterm="",
            )
        )
        lexer = "diff"
        line_numbers = False
    else:
        to_display = fix_suggestion.formatted_code
        lexer = language
        line_numbers = True

    syntax = Syntax(
        to_display,
        lexer,
        theme=theme,
        line_numbers=line_numbers,
    )
    console.print(Panel(syntax))
    console.print()


def prompt_accept_fix(console: Console | None = None) -> bool:
    if console is None:
        console = get_rich_console()

    return typer.confirm("Accept this fix?", default=True)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum

from pydantic import BaseModel

from .openapi import KernelLanguage


class TargetDevice(Enum):
    H100 = "H100"
    H200 = "H200"
    B200 = "B200"
    L40S = "L40S"
    MI300X = "MI300X"
    ADRENO_830 = "Adreno 830"
    ADRENO_750 = "Adreno 750"
    HEXAGON_V79 = "Hexagon v79"
    HEXAGON_V75 = "Hexagon v75"

    def get_default_language(self) -> KernelLanguage:
        match self:
            case TargetDevice.H100 | TargetDevice.H200 | TargetDevice.B200 | TargetDevice.L40S:
                return KernelLanguage.cuda
            case TargetDevice.MI300X:
                return KernelLanguage.hip
            case TargetDevice.ADRENO_750 | TargetDevice.ADRENO_830:
                return KernelLanguage.opencl
            case TargetDevice.HEXAGON_V75 | TargetDevice.HEXAGON_V79:
                return KernelLanguage.ripple

    def supports_language(self, lang: KernelLanguage) -> bool:
        match self:
            case TargetDevice.H100 | TargetDevice.H200 | TargetDevice.B200 | TargetDevice.L40S:
                return lang in {
                    KernelLanguage.cuda,
                    KernelLanguage.triton,
                }

            case TargetDevice.MI300X:
                return lang in {
                    KernelLanguage.triton,
                    KernelLanguage.hip,
                }

            case TargetDevice.ADRENO_750 | TargetDevice.ADRENO_830:
                return lang == KernelLanguage.opencl

            case TargetDevice.HEXAGON_V75 | TargetDevice.HEXAGON_V79:
                return lang == KernelLanguage.ripple

    def to_api_device(self) -> str:
        match self:
            case TargetDevice.H100:
                return "nvidia:h100"
            case TargetDevice.H200:
                return "nvidia:h200"
            case TargetDevice.B200:
                return "nvidia:b200"
            case TargetDevice.L40S:
                return "nvidia:L40S"
            case TargetDevice.MI300X:
                return "amd:mi300x"
            case TargetDevice.ADRENO_830:
                return "qualcomm:snapdragon_8_elite_gpu"
            case TargetDevice.ADRENO_750:
                return "qualcomm:snapdragon_8_gen_3_gpu"
            case TargetDevice.HEXAGON_V79:
                return "qualcomm:snapdragon_8_elite_npu"
            case TargetDevice.HEXAGON_V75:
                return "qualcomm:snapdragon_8_gen_3_npu"

    @classmethod
    def from_api_name(cls, name: str) -> "TargetDevice":
        match name:
            case "nvidia:h100":
                return cls.H100
            case "nvidia:h200":
                return cls.H200
            case "nvidia:b200":
                return cls.B200
            case "nvidia:L40S":
                return cls.L40S
            case "amd:mi300x":
                return cls.MI300X
            case "qualcomm:snapdragon_8_elite_gpu":
                return cls.ADRENO_830
            case "qualcomm:snapdragon_8_gen_3_gpu":
                return cls.ADRENO_750
            case "qualcomm:snapdragon_8_elite_npu":
                return cls.HEXAGON_V79
            case "qualcomm:snapdragon_8_gen_3_npu":
                return cls.HEXAGON_V75
            case _:
                raise NotImplementedError()


class SessionExtra(BaseModel):
    speedup: float | None = None
    device: TargetDevice | None = None



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import textwrap
from uuid import UUID

from .errors import Http404, HttpError
from .conn import Connection
from .auth import get_current_credentials
from ..models.internal import TargetDevice, SessionExtra
from ..models.openapi import (
    KernelLanguage,
    PredefinedKernelGenerationRequest,
    AgentSession,
    ThinkingLevel,
    AgentGenerationAttempt,
    AgentSessions,
    AgentSessionSummary,
    SessionKernels,
    EvaluatedKernel,
    UserInstruction,
)


async def new_session(
    conn: Connection,
    problem_id: UUID,
    language: KernelLanguage,
    device: TargetDevice,
    label: str,
    atol: float,
    rtol: float,
    user_prompt: str,
) -> str:
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    req = PredefinedKernelGenerationRequest(
        label=label,
        problem_id=str(problem_id),
        thinking_level=ThinkingLevel.high,
        backend=language,
        target_hardware=device.to_api_device(),
        user_prompt=user_prompt,
        budget_limit=None,
        atol=atol,
        rtol=rtol,
    )

    repl = await conn.post("agent-session", req, reply_format=AgentSession, token=creds.token)
    return str(repl.id)


async def fetch_session_extra(conn: Connection, session_id: UUID) -> SessionExtra | None:
    """Fetch extra session data from best-attempt endpoint."""
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    try:
        repl = await conn.get(
            f"agent-session/{session_id}/best-attempt",
            reply_format=AgentGenerationAttempt,
            token=creds.token,
        )
    except Http404:
        return None

    device = repl.request.target_hardware
    bench = repl.evaluation_state.benchmarking_result
    speedup: float | None = None
    if bench is not None:
        ref_compiled = bench.ref_compiled_time
        optimized = bench.optimized_time
        if ref_compiled and optimized and optimized > 0:
            speedup = ref_compiled / optimized

    return SessionExtra(
        speedup=speedup,
        device=TargetDevice.from_api_name(device) if device else None,
    )


async def resolve_session(sessions: list[AgentSessionSummary], session_id: str) -> AgentSessionSummary | None:
    """Find session matching the given ID prefix."""
    matches: list[AgentSessionSummary] = []
    for s in sessions:
        if str(s.id).startswith(session_id):
            matches.append(s)

    if not matches:
        return None
    elif len(matches) > 1:
        session_list = [" * " + str(s.id) for s in matches]
        session_block = textwrap.indent("\n".join(session_list), "    ")
        raise ValueError(f"Session ID prefix: {session_id!r} is matching more than one session:\n" + session_block)

    return matches[0]


async def get_user_sessions(conn: Connection) -> list[AgentSessionSummary]:
    """Fetch a list of sessions belonging to the current user."""
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    ret: list[AgentSessionSummary] = []

    offset = 0
    while True:
        repl = await conn.get(f"agent-session?offset={offset}", reply_format=AgentSessions, token=creds.token)
        for s in repl.sessions:
            if not s.deleted_at:
                ret.append(s)

        offset += len(repl.sessions)
        if offset >= repl.total:
            break

    return ret


async def get_session_kernels(conn: Connection, session_id: str) -> list[list[EvaluatedKernel]]:
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    ret: list[list[EvaluatedKernel]] = []

    repl = await conn.get(
        f"agent-session/{session_id}/kernels",
        reply_format=SessionKernels,
        token=creds.token,
    )

    sorted_attempts = sorted(repl.attempts, key=lambda a: a.attempt_number)
    for attempt in sorted_attempts:
        if not attempt.kernels:
            continue

        ret.append(attempt.kernels)

    return ret


async def stop_instruction(conn: Connection, instruction_id: str) -> UserInstruction:
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    repl = await conn.post(
        f"agent-session/instruction/{instruction_id}/stop",
        reply_format=UserInstruction,
        token=creds.token,
    )
    return repl


async def stop_job(conn: Connection, session_id: UUID) -> bool:
    session = await get_session(conn, str(session_id))
    for attempt in session.generation_attempts:
        if attempt.user_instruction and attempt.stop_requested_at is None:
            try:
                await stop_instruction(conn, str(attempt.user_instruction.id))
                return True
            except HttpError:
                pass
    return False


async def get_session(conn: Connection, session_id: str) -> AgentSession:
    creds = get_current_credentials()
    if creds is None:
        raise RuntimeError("User needs to be logged in")

    repl = await conn.get(
        f"agent-session/{session_id}",
        reply_format=AgentSession,
        token=creds.token,
    )
    return repl



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .utils import (
    static_property,
    add_module_properties,
)


def _get_version() -> str:
    from . import version

    return version.version


def _get_has_repo() -> bool:
    from . import version

    return version.has_repo


def _get_repo() -> str:
    from . import version

    return version.repo


def _get_commit() -> str:
    from . import version

    return version.commit


__version__: str
__has_repo__: bool
__repo__: str
__commit__: str


__all__ = [
    "__version__",
    "__has_repo__",
    "__repo__",
    "__commit__",
]


add_module_properties(
    __name__,
    {
        "__version__": static_property(staticmethod(_get_version)),
        "__has_repo__": static_property(staticmethod(_get_has_repo)),
        "__repo__": static_property(staticmethod(_get_repo)),
        "__commit__": static_property(staticmethod(_get_commit)),
    },
)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import types
from typing import Any, Callable, Iterable, TYPE_CHECKING, Protocol, overload, TypeVar
from typing_extensions import Self
from types import TracebackType, EllipsisType
from functools import lru_cache

from rich.console import Console


U = TypeVar("U", covariant=True)
V = TypeVar("V", covariant=False)


class static_property(property):
    def __init__(
        self,
        fget: Callable[[], Any] | None = None,
        fset: Callable[[Any], None] | None = None,
        fdel: Callable[[], None] | None = None,
        doc: str | None = None,
    ) -> None:
        if fget is not None and not isinstance(fget, staticmethod):
            fget = staticmethod(fget)
        if fset is not None and not isinstance(fset, staticmethod):
            fset = staticmethod(fset)
        if fdel is not None and not isinstance(fdel, staticmethod):
            fdel = staticmethod(fdel)
        super().__init__(fget, fset, fdel, doc)  # type: ignore

    def __get__(self, inst: Any, cls: type | None = None) -> Any:
        if inst is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget.__get__(inst, cls)()  # pylint: disable=no-member

    def __set__(self, inst: Any, val: Any) -> None:
        if self.fset is None:
            raise AttributeError("can't set attribute")

        # pylint: disable=no-member
        return self.fset.__get__(inst)(val)  # type: ignore

    def __delete__(self, inst: Any) -> None:
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        # pylint: disable=no-member
        return self.fdel.__get__(inst)()  # type: ignore


class LazyModuleType(types.ModuleType):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    def __getattribute__(self, name: str) -> Any:
        _props = super().__getattribute__("_props")
        if name in _props:
            return object.__getattribute__(self, name)
        else:
            return types.ModuleType.__getattribute__(self, name)

    def __dir__(self) -> Iterable[str]:
        ret = super().__dir__()
        ret.extend(self._props)  # type: ignore
        return ret


def add_module_properties(module_name: str, properties: dict[str, Any]) -> None:
    module = sys.modules[module_name]
    replace = False
    if isinstance(module, LazyModuleType):
        hacked_type = type(module)
    else:
        hacked_type = type(
            "LazyModuleType__{}".format(module_name.replace(".", "_")),
            (LazyModuleType,),
            {"_props": set()},
        )
        replace = True

    for name, prop in properties.items():
        if not isinstance(prop, property):
            prop = property(prop)
        setattr(hacked_type, name, prop)
        hacked_type._props.add(name)  # type: ignore

    if replace:
        new_module = hacked_type(module_name)
        spec = getattr(module, "__spec__", None)
        module.__class__ = new_module.__class__
        module.__name__ = new_module.__name__
        module.__dict__.update(new_module.__dict__)
        module.__spec__ = spec


_env_vars: dict[str, "EnvVar"] = {}


class EnvVarMeta(type):
    def __call__(self, var_name: str, default: str | None = None, hidden: bool = False) -> "EnvVar":
        current = _env_vars.get(var_name)
        if current is not None:
            if current.default == default:
                return current
            raise RuntimeError(
                f"Environment variable with name: {var_name!r} already "
                f"registered with default value of: {current.default!r}"
            )

        new = super().__call__(var_name, default, hidden)
        assert isinstance(new, EnvVar)
        _env_vars[var_name] = new
        return new


class EnvVar(metaclass=EnvVarMeta):
    def __init__(self, var_name: str, default: str | None, hidden: bool = False) -> None:
        self.var_name = var_name
        self.default = default
        self.hidden = hidden
        self._resolved = False
        self._value: str | None = None

    @property
    def value(self) -> str:
        if not self._resolved:
            self._value = os.environ.get(self.var_name, self.default or "")
            self._resolved = True

        assert isinstance(self._value, str)
        return self._value


def get_env_vars() -> Iterable[EnvVar]:
    return sorted(_env_vars.values(), key=lambda e: e.var_name)


NO_RICH = EnvVar("MAKORA_NO_RICH", "")


@lru_cache(maxsize=1, typed=False)
def get_rich_console() -> Console:
    if bool(NO_RICH.value):
        return Console(
            color_system=None,
            force_terminal=False,
            force_interactive=False,
            no_color=True,
            tab_size=4,
            highlight=False,
            emoji=False,
        )

    return Console(
        tab_size=4,
    )


class _dummy_context:
    def __init__(self, value: Any = ...) -> None:
        self.value = value

    def __enter__(self) -> Any:
        if self.value is Ellipsis:
            return self
        return self.value

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        return None


class _async_dummy_context:
    def __init__(self, value: Any = ...) -> None:
        self.value = value

    async def __aenter__(self) -> Any:
        if self.value is Ellipsis:
            return self
        return self.value

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        return None


if TYPE_CHECKING:

    class SelfContext(Protocol):
        def __enter__(self) -> Self: ...

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> bool | None: ...

    class ValueContext(Protocol[U]):
        def __enter__(self) -> U: ...

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> bool | None: ...

    class SelfAsyncContext(Protocol):
        async def __aenter__(self) -> Self: ...

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> bool | None: ...

    class ValueAsyncContext(Protocol[U]):
        async def __aenter__(self) -> U: ...

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> bool | None: ...

    @overload
    def dummy_context(value: EllipsisType = ...) -> SelfContext: ...

    @overload
    def dummy_context(value: V) -> ValueContext[V]: ...

    def dummy_context(value: Any = ...) -> SelfContext | ValueContext[Any]:
        return _dummy_context(value)

    @overload
    def async_dummy_context(value: EllipsisType = ...) -> SelfAsyncContext: ...

    @overload
    def async_dummy_context(value: V) -> ValueAsyncContext[V]: ...

    def async_dummy_context(
        value: Any = ...,
    ) -> SelfAsyncContext | ValueAsyncContext[Any]:
        return _async_dummy_context(value)
else:

    def dummy_context(value=...):
        return _dummy_context(value)

    def async_dummy_context(value=...):
        return _async_dummy_context(value)



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rich.console import Console

from ..utils import get_rich_console


MAKORA_LOGO = """[bold cyan]
 ███╗   ███╗ █████╗ ██╗  ██╗ ██████╗ ██████╗  █████╗ 
 ████╗ ████║██╔══██╗██║ ██╔╝██╔═══██╗██╔══██╗██╔══██╗
 ██╔████╔██║███████║█████╔╝ ██║   ██║██████╔╝███████║
 ██║╚██╔╝██║██╔══██║██╔═██╗ ██║   ██║██╔══██╗██╔══██║
 ██║ ╚═╝ ██║██║  ██║██║  ██╗╚██████╔╝██║  ██║██║  ██║
 ╚═╝     ╚═╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝ ╚═╝  ╚═╝╚═╝  ╚═╝[/bold cyan]
[dim]        Automatically unlock peak GPU performance.[/dim]"""


MAKORA_MINI = "[bold cyan]MAKORA[/bold cyan]"


def print_header(console: Console | None = None) -> None:
    """Print the MAKORA header"""
    if console is None:
        console = get_rich_console()

    console.print(MAKORA_LOGO)
    console.print()


def print_mini_header(subtitle: str = "", console: Console | None = None) -> None:
    """Print a minimal header for commands"""
    if console is None:
        console = get_rich_console()

    text = f"{MAKORA_MINI}"
    if subtitle:
        text += f" [dim]|[/dim] [white]{subtitle}[/white]"
    console.print(text)
    console.print()



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Claude Code plugin installation utilities."""

import shutil
import subprocess
from pathlib import Path

PLUGIN_DIR = Path(__file__).resolve().parent.parent / "data" / "claude-plugin"


class ClaudeInstallError(Exception):
    """Raised when Claude Code plugin installation fails."""

    pass


def install_claude_code_plugin() -> dict[str, str]:
    """Install the Makora plugin for Claude Code."""
    messages: dict[str, str] = {}
    data_path = PLUGIN_DIR.resolve()

    cache_path = Path.home() / ".claude" / "plugins" / "cache" / "makora"
    if cache_path.exists():
        shutil.rmtree(cache_path)
        messages["cache_removed"] = f"Removed cache folder: {cache_path}"

    result = subprocess.run(
        ["claude", "plugin", "marketplace", "remove", "makora"],
        capture_output=True,
        text=True,
    )
    if result.returncode == 0:
        messages["removed"] = "Removed existing makora marketplace."

    result = subprocess.run(
        ["claude", "plugin", "marketplace", "add", str(data_path)],
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        raise ClaudeInstallError(f"Failed to add marketplace: {result.stderr}")
    messages["marketplace"] = result.stdout.strip() if result.stdout.strip() else "Marketplace added."

    result = subprocess.run(
        ["claude", "plugin", "install", "makora-plugin@makora"],
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        raise ClaudeInstallError(f"Failed to install plugin: {result.stderr}")
    messages["plugin"] = result.stdout.strip() if result.stdout.strip() else "Plugin installed."

    return messages



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any

version = "1.0.0"
repo = "unknown"
commit = "unknown"
has_repo = False

try:
    import git
    from pathlib import Path

    try:
        r = git.Repo(Path(__file__).parents[1])
        has_repo = True

        if not r.remotes:
            repo = "local"
        else:
            repo = r.remotes.origin.url

        commit = r.head.commit.hexsha  # cSpell: disable-line
        status = []
        if r.is_dirty():
            status.append("dirty")
        if r.untracked_files:
            status.append(f"+{len(r.untracked_files)} untracked")
        if status:
            commit += f" ({','.join(status)})"
    except git.InvalidGitRepositoryError:
        raise ImportError()
except ImportError:
    pass

try:
    import importlib.util
    from pathlib import Path

    _dist_info_file = Path(__file__).parent.joinpath("_dist_info.py")
    if _dist_info_file.exists():
        _spec = importlib.util.spec_from_file_location("_dist_info", _dist_info_file)
        assert _spec is not None
        _dist_info = importlib.util.module_from_spec(_spec)
        assert _dist_info is not None
        assert _spec.loader is not None
        _spec.loader.exec_module(_dist_info)
        assert not has_repo, "_dist_info should not exist when repo is in place"
        assert version == _dist_info.version
        repo = _dist_info.repo
        commit = _dist_info.commit
except (ImportError, SystemError):
    pass


def info() -> dict[str, Any]:
    g = globals()
    return {k: g[k] for k in __all__}


__all__ = ["version", "repo", "commit", "has_repo"]



# Copyright 2026 Makora Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from pathlib import Path
from typing import Annotated

import typer

from ..models.openapi import ExpertGenerateRequest, KernelGenerationResult, KernelLanguage
from ..models.internal import TargetDevice
from ..web.auth import ensure_authenticated, get_current_credentials
from ..web.conn import open_connection


async def cli_expert_generate_async(
    file: Path,
    problem: Path | None,
    device: TargetDevice,
    language: str,
    speedup: float | None,
    url: str | None = None,
) -> None:
    creds = get_current_credentials()
    if creds is None:
        raise SystemExit("You need to login first with 'makora login'")

    if not file.exists():
        typer.echo(f"Error: File not found: {file}", err=True)
        raise typer.Exit(1)
    if problem is not None and not problem.exists():
        typer.echo(f"Error: File not found: {problem}", err=True)
        raise typer.Exit(1)

    kernel_code = file.read_text()
    problem_code = problem.read_text() if problem is not None else None

    try:
        lang = KernelLanguage(language)
    except ValueError:
        typer.echo(
            f"Error: Invalid language '{language}'. Must be one of: {[e.value for e in KernelLanguage]}",
            err=True,
        )
        raise typer.Exit(1) from None

    target_hardware = device.to_api_device()

    request = ExpertGenerateRequest(
        kernel_code=kernel_code,
        reference_code=problem_code,
        language=lang,
        target_hardware=target_hardware,
        current_speedup=speedup,
        benchmark_info=None,
    )

    typer.echo("Generating optimized kernel...", err=True)

    async with open_connection(url) as conn:
        await ensure_authenticated(conn)
        response = await conn.post(
            "additional-tools/expert-generate",
            request,
            reply_format=KernelGenerationResult,
            token=creds.token,
        )

    if response.summary:
        typer.echo(f"Summary: {response.summary}", err=True)

    typer.echo(response.code)


def cli_expert_generate(
    file: Annotated[Path, typer.Argument(help="Path to the kernel file to optimize")],
    device: Annotated[TargetDevice, typer.Option("-d", "--device", help="Device type")],
    problem: Annotated[
        Path | None,
        typer.Option("-p", "--problem", help="Path to the reference/problem file for context"),
    ] = None,
    language: Annotated[str, typer.Option("-l", "--language", help="Programming language")] = "cuda",
    speedup: Annotated[
        float | None,
        typer.Option("--speedup", help="Current speedup vs baseline for context"),
    ] = None,
    url: Annotated[str | None, typer.Option(help="Override the Makora API URL")] = None,
) -> None:
    """Generate an optimized GPU kernel using expert optimization patterns."""
    try:
        asyncio.run(
            cli_expert_generate_async(
                file=file,
                problem=problem,
                device=device,
                language=language,
                speedup=speedup,
                url=url,
            )
        )
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1) from e
